{
=============================================================================
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  |
| AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    |
| IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   |
| ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR  |
| ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL       |
| DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR   |
| SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER   |
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT           |
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY    |
| OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  |
| DAMAGE.                                                                      |
=============================================================================
  Auteur: D.J.C. Blijker
  Datum:  24-4-2003
  Versie: 2.0

  Project: Log-Daemon
  Module: LogDaemon_Logic
=============================================================================
  This unit will work in Delphi as well as in Kylix!!!
=============================================================================
}
unit LogDaemon_LogicLayer;

interface

// UNDEF (or remove) this line if you don't want service-provider support..
{$DEFINE USE_SERVICE_PROVIDER}

// By default, connections to this logdaemon can be made both by the
// IPC-protocol and the UDP-protocol. But you can create a 'lean-and-mean'
// version without IPC-support or without UDP-support (or without both but
// I can't imagine that would be usefull ;-)

{$DEFINE USE_IPC_SERVER}
{$DEFINE USE_UDP_SERVER}

uses
  SysUtils, IniFiles, Forms, extctrls, classes,
  LogDaemon_LogFileLayer, LogDaemon_Common,
{$IFDEF USE_SERVICE_PROVIDER}
  ServiceProviderLayer, ServiceBroker_Shared,
{$ENDIF}
  HxDataExchange, HxTcpIpcUdp;

const
  ALIVE_TIMEOUT = 20;

type
  TConnectionType = ( ctIPC, ctUDP );

  TClientInfo = record
    iDescr: integer;
    mmsMem: TMemoryStream;
    sExeName: string;
    sIP: string;
    sPortNr: string;
    sLogFileName: string;   // Base logfile name..
    eConnectionType: TConnectionType;
    tmrDump: TTimer;
    iInterval: integer;
    iStdInterval: integer;  // in Seconds
    iAliveTimeout: integer; // in seconds..
    lfLogFile: TDaemonLogFile;
  end;
  PClientInfo = ^TClientInfo;

  TClientInfoList = class( TList )
  protected
    procedure Put( index: integer; pElement: PClientInfo );
    function  Get( index: integer ): PClientInfo;
    procedure Notify( Ptr: Pointer; Action: TListNotification); override;

  public
    property items[ index: integer]: PClientInfo read Get
             write Put; default;
    procedure Add( ptr: PClientInfo );
  end;

  TMessNotify = procedure( sMess: string )of object;

  TProc = procedure( sRootText, sClient: string )of object;

  TServerDaemon = class
  private
    { Private declarations }
    FOnNewClient: TMessNotify;
    FOnClientAborted: TMessNotify;
    FOnMessage: TMessNotify;
{$IFDEF USE_IPC_SERVER}
    svrIpc: THxIpcServer;
{$ENDIF}
{$IFDEF USE_UDP_SERVER}
    udpSvr: THxUdp;
{$ENDIF}
{$IFDEF USE_SERVICE_PROVIDER}
    spServProvider: THxServiceProvider;
{$ENDIF}
    iDefIntervalTime: integer;
    lstClients: TClientInfoList;
    deDataEx: THxDataExchange;
    function  GetClientIndexByTimer(tmr: TTimer): integer;
    function  CreateDefClientLogFileName(iIdx: integer): string;
    procedure Notify( sMess: string );
    procedure HandleTimer( Sender: TObject );
    procedure HandleString( ptr: pointer; iSize, iIdx: integer );
    procedure HandleWarning( ptr: pointer; iSize, iIdx: integer );
    procedure DumpToFile( iIdx: integer );
    procedure SetIntervalTime( iValue: integer );
    procedure HandleIdxClientAborted( iIdx: integer );
    function  HandleData( ptr: pointer; iType, iSize, iIdx: integer ): boolean;
// Handle 'commands' from the client..
    procedure HandleApplicationName( ptr: pointer; iIdx: integer );
    procedure HandleStartNewFile( ptr: pointer; iIdx: integer );
    procedure HandleNamedLog( ptr: pointer; iIdx: integer );
    procedure HandleSetNrOfFiles( ptr: pointer; iIdx: integer );
    procedure HandleSetSwitchTime( ptr: pointer; iIdx: integer );
{$IFDEF USE_IPC_SERVER}
    function  GetClientIndexByDescr( iDescr: integer ): integer;
    procedure HandleIpcClientAborted( iDescr: integer );
    procedure HandleIpcNewClient( iClientDescr: integer );
    procedure HandleIpcData( ptr: pointer; iSize, iDescr: integer );
{$ENDIF}
{$IFDEF USE_UDP_SERVER}
    function  GetClientIndexByIpAndPort( sIP_Sender, sPort: string ): integer;
    procedure HandleUdpData( ptr: pchar; iSize: integer;
                             sIP_Sender, sPort: string );
    procedure HandleNewUdpClient( sIP_Sender, sPort: string );
    procedure HandleClientAlive( iIdx: integer );
    procedure SendAbortMessage( sIP_Dest, sPort: string );
    procedure SendCommand( sIP_Dest, sPort: string; iCmd: integer );
    procedure SendAliveRequest( sIP_Dest, sPort: string );
    procedure SendAcknowledge( sIP_Dest, sPort: string );
    procedure HandleUdpClientAborted(iIdx: integer);
{$ENDIF}
  public
    constructor Create; 
    destructor  Destroy; override;
    property  OnNewClient: TMessNotify read FOnNewClient write FOnNewClient;
    property  OnClientAborted: TMessNotify read FOnClientAborted
              write FOnClientAborted;
    property  OnMessage: TMessNotify read FOnMessage write FOnMessage;
    property  DefaultIntervalTime: integer read iDefIntervalTime
              write SetIntervalTime;
    procedure DumpAllToFile;
    procedure GetConnectedClients( sRootText: string; prcAddClient: TProc );
  end;

implementation

{ TClientInfoList }

procedure TClientInfoList.Add( ptr: PClientInfo );
begin
  inherited Add( ptr );
end;

function TClientInfoList.Get( index: integer ): PClientInfo;
begin
  result := PClientInfo( inherited Get( index ));
end;

procedure TClientInfoList.Notify( Ptr: Pointer; Action: TListNotification );
begin
  if Action = lnDeleted then
  begin
    PClientInfo( Ptr )^.mmsMem.Free;
    PClientInfo( Ptr )^.tmrDump.Free;
    PClientInfo( Ptr )^.lfLogFile.Free;
    Dispose( ptr );
  end;
end;

procedure TClientInfoList.Put( index: integer; pElement: PClientInfo );
begin
  inherited Put( index, pElement );
end;

{ TServerDaemon }

constructor TServerDaemon.Create;
begin
{$IFDEF USE_SERVICE_PROVIDER}
  spServProvider := THxServiceProvider.Create;
  {$IFDEF USE_IPC_SERVER}
  spServProvider.ProvideService( LOGDAEMON_SERVICE, LOGDAEMON_PORT, pcIpc );
  {$ENDIF}
  {$IFDEF USE_UDP_SERVER}
  spServProvider.ProvideService( LOGDAEMON_SERVICE, LOGDAEMON_PORT, pcUdp );
  {$ENDIF}
{$ENDIF}

// Create a list for all client connection-data
  lstClients := TClientInfoList.Create;

// Usefull for (un)packing of messages
  deDataEx := THxDataExchange.Create;

// The default time in seconds between write-to-file actions..
  iDefIntervalTime := 10;

{$IFDEF USE_IPC_SERVER}
  svrIpc := THxIpcServer.Create;
  with svrIpc do
  begin
    Port := LOGDAEMON_PORT;
    OnDataAvailable := HandleIpcData;
    OnNewClient := HandleIpcNewClient;
    OnClientAborted := HandleIpcClientAborted;
    Listen;
  end;
{$ENDIF}

{$IFDEF USE_UDP_SERVER}
  udpSvr := THxUdp.Create;
  with udpSvr do
  begin
    ListeningPort := LOGDAEMON_PORT;
    OnDataAvailable := HandleUdpData;
  end;
{$ENDIF}
end;

// Cleanup..
destructor TServerDaemon.Destroy;
begin
  DumpAllToFile;

{$IFDEF USE_SERVICE_PROVIDER}
  spServProvider.Free;
{$ENDIF}

  lstClients.Free;
  deDataEx.Free;

{$IFDEF USE_IPC_SERVER}
  svrIpc.Free;
{$ENDIF}
{$IFDEF USE_UDP_SERVER}
  udpSvr.Free;
{$ENDIF}

  inherited;
end;

// This function translates an IPC-secriptor to an index in the list of
// client-connections.
{$IFDEF USE_IPC_SERVER}
function TServerDaemon.GetClientIndexByDescr( iDescr: integer ): integer;
var
  i: integer;
begin
  result := -1;
  for i := 0 to lstClients.Count-1 do
  begin
    if lstClients[i]^.iDescr = iDescr then
    begin
      result := i;
      break;
    end;
  end;
end;
{$ENDIF}

// This function will look-up the index in the list of client-connections
// by the supplied timer..
function TServerDaemon.GetClientIndexByTimer( tmr: TTimer ): integer;
var
  i: integer;
begin
  result := -1;
  for i := 0 to lstClients.Count-1 do
  begin
    if lstClients[i]^.tmrDump = tmr then
    begin
      result := i;
      break;
    end;
  end;
end;

// This function will lookup the index in the list of client-connections
// by IP-address and port number.
{$IFDEF USE_UDP_SERVER}
function TServerDaemon.GetClientIndexByIpAndPort( sIP_Sender,
                                                  sPort: string ): integer;
var
  i: integer;
begin
  result := -1;
  for i := 0 to lstClients.Count-1 do
  begin
    if( lstClients[i]^.sIP = sIP_Sender )and
    ( lstClients[i]^.sPortNr = sPort )then
    begin
      result := i;
      break;
    end;
  end;
end;
{$ENDIF}

procedure TServerDaemon.Notify( sMess: string );
begin
  if Assigned( FOnMessage )then
    FOnMessage( sMess );
end;

// This function will returns a default logfilename. This name will be formed
// by the (clients)applicationname.
function TServerDaemon.CreateDefClientLogFileName( iIdx: integer ): string;
var
  iPos: integer;
begin
  result := lstClients[iIdx]^.sExeName;
  result := ExtractFileName( result );
  iPos := Pos('.', result );
  if iPos > 0 then
    result := Copy( result, 1, iPos-1 );
//  Notify( Format('Creating default logfile_name: %s', [result] ));
end;

// This procedure will be called as soon as a new IPC-client is connected.
// As part of the protocol this client is obliged to send a supplimental
// message containing some meta-data about itself.
{$IFDEF USE_IPC_SERVER}
procedure TServerDaemon.HandleIpcNewClient( iClientDescr: integer );
var
  ptr: PClientInfo;
begin
  New( ptr );
  with ptr^ do
  begin
    iDescr := iClientDescr;
    mmsMem := TMemoryStream.Create;
    sExeName := unknown_yet;
    sIP := na;
    sPortNr := na;
    eConnectionType := ctIPC;
    sLogFileName := unknownFileName;
    tmrDump := TTimer.Create( nil );
    tmrDump.OnTimer := HandleTimer;
    tmrDump.Interval := 1000; // Resolution of 1 sec.
    tmrDump.Enabled := True;
    iStdInterval := iDefIntervalTime;
    iInterval := iDefIntervalTime;
    iAliveTimeout := ALIVE_TIMEOUT;
    lfLogFile := nil;
  end;
  lstClients.Add( ptr );
//  Notify( Format('New IpcClient, name: %s, index: %d', [ptr^.sExeName,
//                  lstClients.Count-1] ));
end;
{$ENDIF}

// This procedure will be called as soon as a new UDP client is 'connected'
// Because UDP is not connection oriented this procedure will only be called
// when no connection-data is known about this address.
{$IFDEF USE_UDP_SERVER}
procedure TserverDaemon.HandleNewUdpClient( sIP_Sender, sPort: string );
var
  ptr: PClientInfo;
begin
  New( ptr );
  with ptr^ do
  begin
    iDescr := -1;
    mmsMem := TMemoryStream.Create;
    sExeName := unknown_yet;
    sIP := sIP_Sender;
    sPortNr := sPort;
    eConnectionType := ctUDP;
    sLogFileName := unknownFileName;
    tmrDump := TTimer.Create( nil );
    tmrDump.OnTimer := HandleTimer;
    tmrDump.Interval := 1000; // 1 sec.
    tmrDump.Enabled := True;
    iInterval := iDefIntervalTime;
    iStdInterval := iDefIntervalTime;
    iAliveTimeout := ALIVE_TIMEOUT;
    lfLogFile := nil;
  end;
  lstClients.Add( ptr );
//  Notify( Format('New UdpClient, name: %s, index: %d', [ptr^.sExeName,
//                  lstClients.Count-1] ));
end;
{$ENDIF}

// This procedure will be called when a client withdraws itself. This is
// regardless of the connection-type
procedure TServerDaemon.HandleIdxClientAborted( iIdx: integer );
var
  sMess: string;
begin
  if iIdx <> -1 then
  begin
// Create a last closing-message..
    sMess := 'Client: '+lstClients[iIdx]^.sExeName+' logged off.';
    Notify( Format('Client: %s logged off', [lstClients[iIdx]^.sExeName] ));

// Dump it to the memory-stream..
    HandleString( pchar( sMess ), Length( sMess )+1, iIdx );

// Dump the mem-stream to the logfile..
    DumpToFile( iIdx );

// Remove this entry from the list of clients..
    lstClients.Delete( iIdx );

// 'Tell' the 'upper-layer' about this 'event'..
    if Assigned( FOnClientAborted )then
      FOnClientAborted( sMess );
  end
end;

// This procedure will be called when an IPC-client withdraws itself.
{$IFDEF USE_IPC_SERVER}
procedure TServerDaemon.HandleIpcClientAborted( iDescr: integer );
var
  iIdx: integer;
begin
  iIdx := GetClientIndexByDescr( iDescr );

  HandleIdxClientAborted( iIdx );
end;
{$ENDIF}

// This procedure is beeing called as soon as an UDP-Client terminates.
// Udp-clients are obliged to send a 'termination-message' (This part of
// the protocol is already implemented in the log-client-class.
// If for any reason the client stops responding it will be removed from this
// server also
{$IFDEF USE_UDP_SERVER}
procedure TServerDaemon.HandleUdpClientAborted( iIdx: integer );
begin
  HandleIdxClientAborted( iIdx );
end;
{$ENDIF}

// This procedure will be called as soon some data is send to this server.
// A True is returned if the message is recognized, else a False.
function TServerDaemon.HandleData( ptr: pointer;
                                   iType, iSize, iIdx: integer ): boolean;
begin
  result := True;  // until proven otherwise..
  case iType of
    PLM_APPLICATION_NAME: HandleApplicationName( ptr, iIdx );
    PLM_START_FILE: HandleStartNewFile( ptr, iIdx );
    PLM_NAMED_LOG: HandleNamedLog( ptr, iIdx );
    PLM_NR_OF_FILES: HandleSetNrOfFiles( ptr, iIdx );
    PLM_SWITCHTIME: HandleSetSwitchTime( ptr, iIdx );
{$IFDEF USE_UDP_SERVER}
    PLM_BYEBYE: HandleUdpClientAborted( iIdx );
    PLM_ALIVE_AND_KICKING: HandleClientAlive( iIdx );
{$ENDIF}
    LM_STRING: HandleString( ptr, iSize, iIdx );
    LM_WARNING: HandleWarning( ptr, iSize, iIdx );
  else
    result := False;
  end;
end;

{$IFDEF USE_IPC_SERVER}
// This is the eventhandler for the 'OnData' of the IPC-connection
procedure TServerDaemon.HandleIpcData( ptr: pointer; iSize, iDescr: integer );
var
  iType, iIdx: integer;
begin
  try
    iType := -1;
// Try to read the messagetype
    ptr := deDataEx.ReadInteger( ptr, iType, iSize );

// llokup the rigth client data
    iIdx := GetClientIndexByDescr( iDescr );
    if iIdx = -1 then
    begin
      Notify( Format( unknown_client_descr, [iDescr, TranslateMessageNr( iType ),
                                           iSize] ));
      exit;
    end;

// Handle this message
    if not( HandleData( ptr, iType, iSize, iIdx ))then
      Notify( Format( wrong_ipc_message, [TranslateMessageNr( iType ), iSize] ));

  except
    on E: Exception do
    begin
      Notify( Format( exception_while_handling, [TranslateMessageNr( iType ),
                                              E.Message] ));
    end;
  end;
end;
{$ENDIF}

{$IFDEF USE_UDP_SERVER}
// This is the eventhandler for the 'OnData' of the UDP-connection
procedure TServerDaemon.HandleUdpData( ptr: pchar; iSize: integer;
                                       sIP_Sender, sPort: string );
var
  iType, iIdx: integer;
begin
// Lookup the client data
  iIdx := GetClientIndexByIpAndPort( sIP_Sender, sPort );
  try
    iType := -1;
// Read the message type..
    ptr := deDataEx.ReadInteger( ptr, iType, iSize );

// If this client is not known (yet) than the message MUST begin with an
// 'introduction message'. If not this message will be ignored..
    if iIdx = -1 then
    begin
      if iType <> PLM_APPLICATION_NAME then // We don't want anything else
        exit;                               // Ignore

// Create a new entry in the list of clientdata
      HandleNewUdpClient( sIP_Sender, sPort );
      iIdx := lstClients.Count -1;
// Inform the client a (virtual) connection is setup
      SendAcknowledge( sIP_Sender, sPort );
    end;

// Handle the message
    if not( HandleData( ptr, iType, iSize, iIdx ))then
      Notify( Format( wrong_ipc_message, [TranslateMessageNr( iType ), iSize] ));

  except
    on E: Exception do
    begin
      Notify( Format( exception_while_handling, [TranslateMessageNr( iType ),
                                              E.Message] ));
    end;
  end;
end;
{$ENDIF}

// Udp-clients moeten op regelmatige tijden een 'alive' bevestiging sturen.
// Zij doen dit op 'uitnodiging' van deze server.
// Udpclient are compelled to send 'alive' messages to this server. They do so
// on invitation of this server..
{$IFDEF USE_UDP_SERVER}
procedure TServerDaemon.HandleClientAlive( iIdx: integer );
begin
//  Notify('Reset of alive timeout');
  if( iIdx >= 0 )and( iIdx < lstClients.Count )then
    lstClients[iIdx]^.iAliveTimeout := ALIVE_TIMEOUT;
end;
{$ENDIF}

// This procedure is being called as soon as a client sends his application
// name. This is part of the protocol.
procedure TServerDaemon.HandleApplicationName( ptr: pointer; iIdx: integer );
var
  s, sExeFileName: string;
begin
// Read exe-name..
  deDataEx.ReadString( ptr, sExeFileName );
  with lstClients[iIdx]^ do
  begin
    sExeName := sExeFileName;

    if Assigned( lfLogFile )then
      lfLogFile.Free;

// Default the logfilename will be derived from the application name.
    sLogFileName := CreateDefClientLogFileName( iIdx );
    lfLogFile := TDaemonLogFile.Create( sLogFileName );

// Add the first line in the buffer containing the name and
// connection date-time.
    s := 'Client: ' + sExeName + ' logged on.';
    HandleString( pchar( s ), Length( s )+1, iIdx );
  end;

// Tell interface about this 'event'
  if Assigned( FOnNewClient )then
    FOnNewClient('New client: '+sExeFileName );

  Notify( Format('Client name: %s, index: %d', [sExeFileName, iIdx]));
end;

// This procedure will be called as soon as a client provides a logfilename
// itself. Default is the client-application name with a '.log' extension.
procedure TServerDaemon.HandleNamedLog( ptr: pointer; iIdx: integer );
var
  sLogName: string;
begin
  deDataEx.ReadString( ptr, sLogName );
  with lstClients[iIdx]^ do
  begin
    sLogFileName := sLogName;
    if Assigned( lfLogFile )then
      lfLogFile.FilePathName := sLogName
    else
    begin
      lfLogFile := TDaemonLogFile.Create( sLogName );
      Notify('Warning: LogFileName was unknown while setting the new name: '+
            sLogName );
    end;
  end;
end;

// This procedure is called as soon as a client provides the maximum
// serial number of the filename. Default is 10.
procedure TServerDaemon.HandleSetNrOfFiles( ptr: pointer; iIdx: integer );
var
  iNrOfFiles: integer;
begin
  deDataEx.ReadInteger( ptr, iNrOfFiles );
  with lstClients[iIdx]^ do
  begin
    if Assigned( lfLogFile )then
      lfLogFile.MaxVersionNr := iNrOfFiles
    else
      Notify('Error: LogFileName is unknown while setting the nr of files.');
  end;
end;

// This procedure is called as soon as a client provides the switch-time
// between succeeding serialnumbers.
procedure TServerDaemon.HandleSetSwitchTime( ptr: pointer; iIdx: integer );
var
  iSwitchTime: integer;
begin
  deDataEx.ReadInteger( ptr, iSwitchTime );
  with lstClients[iIdx]^ do
  begin
    if Assigned( lfLogFile )then
      lfLogFile.SwitchTimeInMinutes := iSwitchTime
    else
      Notify('Error: LogFileName is unknown while setting the switch-time.');
  end;
end;

// This procedure is called as soon as a client sends a force-new-logfile
// message.
procedure TServerDaemon.HandleStartNewFile( ptr: pointer; iIdx: integer );
var
  iVersion: integer;
begin
// Eventueel aanwezig gegevens eerst wegschrijven naar de 'oude' logfile..
  DumpToFile( iIdx );

  with lstClients[iIdx]^ do
  begin
// Look up the new version number..
    if Assigned( lfLogFile )then
    begin
      iVersion := lfLogFile.CurrentVersionNr;
      Inc( iVersion );
      if iVersion > lfLogFile.MaxVersionNr then
        iVersion := 1;

      lfLogFile.CurrentVersionNr := iVersion;
    end
    else
      Notify('Error: LogFileName is unknown while setting a new version number.');
  end;
end;

procedure TServerDaemon.SetIntervalTime( iValue: integer );
var
  i: integer;
begin
  for i := 0 to lstClients.Count-1 do
    lstClients[i].iStdInterval := iValue;

  iDefIntervalTime := iValue;
end;

procedure TServerDaemon.HandleString( ptr: pointer; iSize, iIdx: integer );
var
  sTime: string;
begin
  sTime := FormatDateTime('h:n:ss:zzz', Now )+' ';
  lstClients[iIdx].mmsMem.Write( pointer( sTime )^, Length( sTime ));
  lstClients[iIdx].mmsMem.Write( ptr^, iSize );
end;

procedure TServerDaemon.HandleWarning( ptr: pointer; iSize, iIdx: integer );
var
  sWL: string;
  iWarningLevel: integer;
begin
  ptr := deDataEx.ReadInteger( ptr, iWarningLevel, iSize );
  sWl := FormatDateTime('h:n:ss:zzz', Now )+
         Format( sWarningLevel, [iWarningLevel]);

  lstClients[iIdx].mmsMem.Write( pointer( sWl )^, Length( sWl ));
  lstClients[iIdx].mmsMem.Write( ptr^, iSize );
end;

procedure TServerDaemon.DumpToFile( iIdx: integer );
var
  s: string;
  ptr1, ptr2: pchar;
begin
  with lstClients[iIdx]^ do
  begin
// Reset the interval timer..
    iInterval := iStdInterval;

    if mmsMem.Size = 0 then
      exit; // nothing to do..

// Just to be sure it exists
    if not( Assigned( lfLogFile ))then
      lfLogFile := TDaemonLogFile.Create( sLogFileName );

// Open the logfile..
    lfLogFile.FileOpen := True;
    try
      mmsMem.Position := 0;
      ptr1 := mmsMem.Memory;
      ptr2 := mmsMem.Memory;
      Inc( ptr2, mmsMem.Size );
// And pour out the memstream into the logfile..
      while ptr1 < ptr2 do
      begin
        s := StrPas( ptr1 );
        lfLogFile.WriteToFile( s );
        Inc( ptr1, Length( s )+1 );
      end;
    finally
      mmsMem.Clear;
      lfLogFile.FileOpen := False;
    end;
  end;
end;

procedure TServerDaemon.DumpAllToFile;
var
  i: integer;
begin
  for i := 0 to lstClients.Count-1 do
    DumpToFile( i );
end;

procedure TServerDaemon.HandleTimer( Sender: TObject );
var
  iIdx: integer;
begin
  iIdx := GetClientIndexByTimer( TTimer( Sender ));
  if iIdx < 0 then
    exit; // ...??..

  with lstClients[iIdx]^ do
  begin
{$IFDEF USE_UDP_SERVER}
    if eConnectionType = ctUDP then
    begin
      Dec( iAliveTimeout );
      if iAliveTimeout = ALIVE_TIMEOUT div 2 then
      begin
        SendAliveRequest( sIP, sPortNr );
//        Notify(Format('Sending alive-request for clientnr: %d', [iIdx]));
      end
      else if iAliveTimeout <= 0 then
      begin
        Notify( Format('Sending abort message for client: %s', [sExeName]));
        SendAbortMessage( sIP, sPortNr );
        HandleUdpClientAborted( iIdx );
      end;
    end;
{$ENDIF}
    Dec( iInterval );
    if iInterval <= 0 then
      DumpToFile( iIdx );
  end;
end;

{$IFDEF USE_UDP_SERVER}
procedure TServerDaemon.SendAbortMessage( sIP_Dest, sPort: string );
begin
  SendCommand( sIP_Dest, sPort, UDP_YOUR_LOGGED_OFF );
end;
{$ENDIF}

{$IFDEF USE_UDP_SERVER}
procedure TServerDaemon.SendAliveRequest( sIP_Dest, sPort: string );
begin
  SendCommand( sIP_Dest, sPort, UDP_REQ_ALIVE_SIGN );
end;
{$ENDIF}

{$IFDEF USE_UDP_SERVER}
procedure TServerDaemon.SendAcknowledge( sIP_Dest, sPort: string );
begin
  SendCommand( sIP_Dest, sPort, UDP_ACK );
end;
{$ENDIF}

{$IFDEF USE_UDP_SERVER}
procedure TServerDaemon.SendCommand( sIP_Dest, sPort: string; iCmd: integer );
var
  ptr: pchar;
  iSize: integer;
begin
  with deDataEx do
  begin
    StartCreatingMessage;
    AddString( LOG_SERVER_COOKIE );
    AddInteger( iCmd );
    GetCreatedMessage( ptr, iSize );
  end;
  udpSvr.DestinationAddress := sIP_Dest;
  udpSvr.DestinationPort := sPort;
  udpSvr.SendData( ptr, iSize );
end;
{$ENDIF}

procedure TServerDaemon.GetConnectedClients( sRootText: string;
                                             prcAddClient: TProc );
var
  i: integer;
  s: string;
begin
  for i := 0 to lstClients.Count-1 do
  begin
    if lstClients[i]^.eConnectionType = ctUDP then
      s := '<UDP> '+lstClients[i]^.sExeName
    else
      s := '<IPC> '+lstClients[i]^.sExeName;
    prcAddClient( sRootText, s );
  end;
end;

end.




