{
=============================================================================
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  |
| AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    |
| IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   |
| ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR  |
| ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL       |
| DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR   |
| SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER   |
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT           |
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY    |
| OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  |
| DAMAGE.                                                                      |
=============================================================================
  Author:  D.J.C. Blijker.
  Date:    24-4-2003
  version: 2.0
  e-mail:  c.blijker@wxs.nl

  Project: Log-Daemon
  Module: LogDaemon_Main
=============================================================================
  This unit will work in Delphi as well as in Kylix!!!
=============================================================================
  This module contains the interface logic of the LogDaemon application.
  This application is very suitable to be omplemented as a service.
=============================================================================
History:

  version: 1.0  A windows only release, built on top of a different TCP-class.
  version: 2.0  Initial release which uses the THxTcpIpcUdp.pas module.
==============================================================================
}
unit LogDaemon_Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls,
  LogDaemon_LogicLayer, StdCtrls, ExtCtrls;

type
  TfrmMain = class(TForm)
    tvw: TTreeView;
    mmo: TMemo;
    Splitter1: TSplitter;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    dmn: TServerDaemon;
    procedure AddItemToTreeView(sMasterVal, sChildVal: string;
      trv: TTreeView; iImlIndex: integer);
    procedure CallBack_AddClientToTV(sRootText, sClient: string);
    procedure RefreshClientsPage(sMess: string);
    procedure ShowAllClients;
    procedure ShowMessage(sMess: string);
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.DFM}

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  dmn := TServerDaemon.Create;
  with dmn do
  begin
    OnNewClient := RefreshClientsPage;
    OnClientAborted := RefreshClientsPage;
    OnMessage := ShowMessage;
  end;
end;

procedure TfrmMain.ShowMessage( sMess: string );
begin
  mmo.Lines.Add( sMess );
end;

procedure TfrmMain.RefreshClientsPage( sMess: string );
begin
  ShowAllClients;
end;

procedure TfrmMain.ShowAllClients;
begin
  if not Visible then
    exit;
  tvw.Items.Clear;
  AddItemToTreeView('blablah', 'Clients', tvw, 0 );
  dmn.GetConnectedClients( 'Clients', CallBack_AddClientToTV );
  tvw.FullExpand;
end;

procedure TfrmMain.CallBack_AddClientToTV( sRootText, sClient: string );
begin
  AddItemToTreeView( sRootText, sClient, tvw, 1 );
end;

procedure TfrmMain.AddItemToTreeView( sMasterVal, sChildVal: string;
                                      trv: TTreeView; iImlIndex: integer );
var
  ndeMaster, ndeDummy: TTreeNode;
begin
  ndeMaster := trv.Items.GetFirstNode;
  while ndeMaster <> nil do
  begin
    if Pos( sMasterVal, ndeMaster.Text )<> 0 then
      break;
    ndeMaster := ndeMaster.GetNext;
  end;

  if ndeMaster = nil then
    ndeDummy := trv.Items.Add( nil, sChildVal )
  else
    ndeDummy := trv.Items.AddChild( ndeMaster, sChildVal );

  ndeDummy.ImageIndex := iImlIndex;
  ndeDummy.SelectedIndex := iImlIndex;
end;



procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  dmn.Free;
end;

end.
